"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import collections.abc
from isolate.connections.grpc.definitions import common_pb2
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import google.protobuf.struct_pb2
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class BoundFunction(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ENVIRONMENTS_FIELD_NUMBER: builtins.int
    FUNCTION_FIELD_NUMBER: builtins.int
    @property
    def environments(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
        global___EnvironmentDefinition
    ]: ...
    @property
    def function(self) -> common_pb2.SerializedObject: ...
    def __init__(
        self,
        *,
        environments: collections.abc.Iterable[global___EnvironmentDefinition]
        | None = ...,
        function: common_pb2.SerializedObject | None = ...,
    ) -> None: ...
    def HasField(
        self, field_name: typing_extensions.Literal["function", b"function"]
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "environments", b"environments", "function", b"function"
        ],
    ) -> None: ...

global___BoundFunction = BoundFunction

@typing_extensions.final
class EnvironmentDefinition(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    KIND_FIELD_NUMBER: builtins.int
    CONFIGURATION_FIELD_NUMBER: builtins.int
    kind: builtins.str
    """Kind of the isolate environment."""
    @property
    def configuration(self) -> google.protobuf.struct_pb2.Struct:
        """A free-form definition of environment properties."""
    def __init__(
        self,
        *,
        kind: builtins.str = ...,
        configuration: google.protobuf.struct_pb2.Struct | None = ...,
    ) -> None: ...
    def HasField(
        self, field_name: typing_extensions.Literal["configuration", b"configuration"]
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "configuration", b"configuration", "kind", b"kind"
        ],
    ) -> None: ...

global___EnvironmentDefinition = EnvironmentDefinition



from __future__ import annotations

import shutil
import tempfile
from contextlib import contextmanager
from dataclasses import dataclass, replace
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Iterator

from platformdirs import user_cache_dir

from isolate.backends.common import lock_build_path
from isolate.logs import Log

if TYPE_CHECKING:
    from isolate.backends import BaseEnvironment

_SYSTEM_TEMP_DIR = Path(tempfile.gettempdir())


@dataclass(frozen=True)
class IsolateSettings:
    cache_dir: Path = Path(user_cache_dir("isolate", "isolate"))
    serialization_method: str = "pickle"
    log_hook: Callable[[Log], None] = print

    def log(self, log: Log) -> None:
        self.log_hook(log)

    def _get_temp_base(self) -> Path:
        """Return the base path for creating temporary files/directories.

        If the isolate cache directory is in a different device than the
        system temp base (e.g. /tmp), then it will return a new directory
        under the cache directory."""

        cache_stat = self.cache_dir.stat()
        system_stat = _SYSTEM_TEMP_DIR.stat()
        if cache_stat.st_dev == system_stat.st_dev:
            return _SYSTEM_TEMP_DIR

        if _SYSTEM_TEMP_DIR.samefile(self.cache_dir):
            path = _SYSTEM_TEMP_DIR / "isolate"
        else:
            # This is quite important since if we have a shared cache
            # disk, then /tmp is going to be in a different disk than
            # the cache directory, which would make it impossible to
            # rename() atomically.
            path = self.cache_dir / "tmp"

        path.mkdir(exist_ok=True, parents=True)
        return path

    def _get_lock_dir(self) -> Path:
        """Return a directory which can be used for storing file-based locks."""
        lock_dir = self._get_temp_base() / "locks"
        lock_dir.mkdir(exist_ok=True, parents=True)
        return lock_dir

    @contextmanager
    def cache_lock_for(self, path: Path) -> Iterator[Path]:
        """Create a lock for accessing (and operating on) the given path. This
        means whenever the context manager is entered, the path can be freely
        modified and accessed without any other process interfering."""

        with lock_build_path(path, self._get_lock_dir()):
            try:
                yield path
            except BaseException:
                # If anything goes wrong, we have to clean up the
                # directory (we can't leave it as a corrupted build).
                shutil.rmtree(path, ignore_errors=True)
                raise

    def cache_dir_for(self, backend: BaseEnvironment) -> Path:
        """Return a directory which can be used for caching the given
        environment's artifacts."""
        backend_name = backend.BACKEND_NAME
        assert backend_name is not None

        environment_base_path = self.cache_dir / backend_name
        environment_base_path.mkdir(exist_ok=True, parents=True)
        return environment_base_path / backend.key

    replace = replace


DEFAULT_SETTINGS = IsolateSettings()



from __future__ import annotations

import importlib
import os
from concurrent.futures import ThreadPoolExecutor, as_completed
from contextlib import contextmanager
from dataclasses import dataclass, field, replace
from functools import lru_cache, partial
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    TypeVar,
    cast,
)

import importlib_metadata
import rich
from rich.console import Console
from rich.status import Status
from rich.text import Text

import isolate
from isolate.backends import BaseEnvironment
from isolate.backends.settings import IsolateSettings
from isolate.logs import Log, LogLevel, LogSource

ReturnType = TypeVar("ReturnType")

# Whether to enable debug logging or not.
_DEBUG_LOGGING = bool(os.environ.get("ISOLATE_ENABLE_DEBUG_LOGGING", False))

# List of serialization method options. Ordered by their
# priority/preference.
_SERIALIZATION_OPTIONS = ["dill", "cloudpickle"]


@lru_cache(1)
def _decide_default_backend():
    for option in _SERIALIZATION_OPTIONS:
        try:
            importlib.import_module(option)
        except ImportError:
            continue
        else:
            return option

    rich.print(
        "Falling back to the default serialization method: 'pickle'.\n"
        "For the best experience, please install one of the following: "
        f"{', '.join(map(repr, _SERIALIZATION_OPTIONS))}"
    )
    return "pickle"


def Template(kind: str, **config: Any) -> _EnvironmentBuilder:
    """Create a new environment builder for the given kind (it can be virtualenv
    or conda, depending on the flavor of packages you'd like to add). You can also
    pass any configuration options that the backend supports."""

    default_pkgs = []

    default_backend = _decide_default_backend()
    if default_backend != "pickle":
        default_pkgs.append(
            (default_backend, importlib_metadata.version(default_backend))
        )

    if kind == "virtualenv":
        requirements = config.setdefault("requirements", [])
        requirements.extend(f"{name}=={version}" for name, version in default_pkgs)
        return _PackageCollector(kind, config, collect_into=requirements)
    elif kind == "conda":
        packages = config.setdefault("packages", [])
        packages.extend(f"{name}={version}" for name, version in default_pkgs)
        return _PackageCollector(kind, config, collect_into=packages)
    else:
        raise NotImplementedError(f"Unknown environment kind: {kind}")


@dataclass
class _EnvironmentBuilder:
    def get_definition(self) -> Tuple[Dict[str, Any], IsolateSettings]:
        """Return the isolate definition for this environment!"""
        raise NotImplementedError

    def __rshift__(self, left: Any) -> BoxedEnvironment:
        """Put an environment into a box."""
        if not isinstance(left, Box):
            return NotImplemented

        definition, settings = self.get_definition()
        return left.wrap(definition, settings)

    def _not_supported(self, *args: Any, **kwargs: Any) -> None:
        raise ValueError(
            "Can't run a function on an environment template!"
            "Be sure to forward it into a box first. Like: `environment = template >> box`"
        )

    run = _not_supported
    map = _not_supported


@dataclass(repr=False)
class _PackageCollector(_EnvironmentBuilder):
    kind: str
    config_options: Dict[str, Any]
    collect_into: List[str]

    def __lshift__(self, right: Any) -> Any:
        if not isinstance(right, str):
            return NotImplemented

        self.collect_into.append(right)
        return self

    def get_definition(self) -> Tuple[Dict[str, Any], IsolateSettings]:
        settings = IsolateSettings(serialization_method=_decide_default_backend())
        return {
            "kind": self.kind,
            **self.config_options,
        }, settings

    def __repr__(self):
        return f"{self.kind}({', '.join(f'{k}={v!r}' for k, v in self.config_options.items())})"


@dataclass
class Box:
    """Some sort of a box/machine to run Python on."""

    def wrap(
        self,
        definition: Dict[str, Any],
        settings: IsolateSettings,
    ) -> BoxedEnvironment:
        raise NotImplementedError

    replace = replace


@dataclass
class LocalBox(Box):
    """Run locally."""

    pool_size: int = 1

    def wrap(
        self,
        definition: Dict[str, Any],
        settings: IsolateSettings,
    ) -> BoxedEnvironment:
        return BoxedEnvironment(
            isolate.prepare_environment(
                **definition,
                context=settings,
            ),
            pool_size=self.pool_size,
        )

    def __mul__(self, right: int) -> LocalBox:
        if not isinstance(right, int):
            return NotImplemented

        return self.replace(pool_size=self.pool_size * right)


@dataclass
class RemoteBox(Box):
    """Run on an hosted isolate server."""

    host: str
    pool_size: int = 1

    def wrap(
        self,
        definition: Dict[str, Any],
        settings: IsolateSettings,
    ) -> BoxedEnvironment:
        definition = definition.copy()

        # Extract the kind of environment to use.
        kind = definition.pop("kind", None)
        assert kind is not None, f"Corrupted definition: {definition}"

        # TODO: We should think of a way to set multiple environment definitions

        target_list = [{"kind": kind, "configuration": definition}]

        # Create a remote environment.
        return BoxedEnvironment(
            isolate.prepare_environment(
                "isolate-server",
                host=self.host,
                target_environments=target_list,
                context=settings,
            ),
            pool_size=self.pool_size,
        )

    def __mul__(self, right: int) -> RemoteBox:
        if not isinstance(right, int):
            return NotImplemented

        return self.replace(pool_size=self.pool_size * right)


@dataclass
class BoxedEnvironment:
    """Environment-in-a-box! A user friendly wrapper around the isolate
    environments!"""

    environment: BaseEnvironment
    pool_size: int = 1
    _console: Console = field(
        default_factory=partial(Console, highlighter=None), repr=False
    )
    _status: Optional[Status] = field(default=None, repr=False)
    _active_pool_size: Optional[str] = field(default=None, repr=False)

    def __post_init__(self):
        existing_settings = self.environment.settings
        new_settings = existing_settings.replace(log_hook=self._rich_log)
        self.environment.apply_settings(new_settings)

    def _update_status(self, from_builder: bool = False) -> None:
        if self._status is not None:
            if from_builder:
                self._status.update("Building the environment...", spinner="clock")
            else:
                if self._active_pool_size:
                    self._status.update(
                        f"Running the isolated tasks {self._active_pool_size}",
                        spinner="runner",
                    )
                else:
                    self._status.update("Running the isolated task", spinner="runner")

    def _rich_log(self, log: Log) -> None:
        self._update_status(from_builder=log.source is LogSource.BUILDER)
        if log.source is LogSource.USER:
            # If the log is originating from user code, then print it
            # as a normal message.
            self._console.print(log.message)
        else:
            # Otherwise sprinkle some colors on it!
            allowed_levels = {
                LogLevel.ERROR: "red",
                LogLevel.WARNING: "yellow",
                LogLevel.INFO: "white",
            }
            if _DEBUG_LOGGING or log.source is LogSource.BUILDER:
                allowed_levels[LogLevel.DEBUG] = "blue"
                allowed_levels[LogLevel.TRACE] = "grey"

            if log.level in allowed_levels:
                level = Text(f"[{log.source}]", style=allowed_levels[log.level])
                self._console.print(level + " " + log.message)

    @contextmanager
    def _status_display(self, message: str) -> Iterator[None]:
        assert self._status is None

        try:
            self._status = self._console.status(message)
            with self._status:
                yield
        finally:
            self._status = None
            self._active_pool_size = None

    def run(
        self,
        func: Callable[..., ReturnType],
        *args: Any,
        **kwargs: Any,
    ) -> ReturnType:
        """Run the given `func` in the environment with the passed arguments."""
        executable = partial(func, *args, **kwargs)
        with self._status_display("Preparing for execution..."):
            with self.environment.connect() as connection:
                return cast(ReturnType, connection.run(executable))

    def map(
        self,
        func: Callable[..., ReturnType],
        *iterables: Iterable[Any],
    ) -> Iterable[ReturnType]:
        """Map the given `func` over the given iterables in parallel. pool_size
        is determined by the originating box."""

        with self._status_display("Preparing for execution..."):
            with ThreadPoolExecutor(max_workers=self.pool_size) as executor:
                with self.environment.connect() as connection:
                    futures = [
                        executor.submit(connection.run, partial(func, *args))
                        for args in zip(*iterables)
                    ]
                    self._active_pool_size = f"0/{len(futures)}"
                    for n, future in enumerate(as_completed(futures), 1):
                        yield cast(ReturnType, future.result())
                        self._active_pool_size = f"{n}/{len(futures)}"
                        self._update_status()



"""
Agent process execution wrapper for handling extended PYTHONPATH.
"""

import os
import runpy
import site
import sys


def load_pth_files() -> None:
    """Each site dir in Python can contain some .pth files, which are
    basically instructions that tell Python to load other stuff. This is
    generally used for editable installations, and just setting PYTHONPATH
    won't make them expand so we need manually process them. Luckily, site
    module can simply take the list of new paths and recognize them.

    https://docs.python.org/3/tutorial/modules.html#the-module-search-path
    """
    python_path = os.getenv("PYTHONPATH")
    if python_path is None:
        return None

    # TODO: The order here is the same as the one that is used for generating the
    # PYTHONPATH. The only problem that might occur is that, on a chain with
    # 3 ore more nodes (A, B, C), if X is installed as an editable package to
    # B and a normal package to C, then C might actually take precedence. This
    # will need to be fixed once we are dealing with more than 2 nodes and editable
    # packages.
    for site_dir in python_path.split(os.pathsep):
        site.addsitedir(site_dir)


def main():
    real_agent, *real_arguments = sys.argv[1:]

    load_pth_files()
    # TODO(feat): implement a check to parse "agent-requires" line and see if
    # all the dependencies are installed.
    sys.argv = [real_agent] + real_arguments
    runpy.run_path(real_agent, run_name="__main__")


if __name__ == "__main__":
    load_pth_files()
    main()



from __future__ import annotations

import functools
import os
import shutil
import subprocess
import tempfile
import yaml
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, ClassVar, Dict, List, Optional

from isolate.backends import BaseEnvironment, EnvironmentCreationError
from isolate.backends.common import active_python, logged_io, sha256_digest_of
from isolate.backends.settings import DEFAULT_SETTINGS, IsolateSettings
from isolate.connections import PythonIPC

# Specify the path where the conda binary might reside in (or
# mamba, if it is the preferred one).
_CONDA_COMMAND = os.environ.get("CONDA_EXE", "conda")
_ISOLATE_CONDA_HOME = os.getenv("ISOLATE_CONDA_HOME")

# Conda accepts the following version specifiers: =, ==, >=, <=, >, <, !=
_CONDA_VERSION_IDENTIFIER_CHARS = (
    "=",
    "<",
    ">",
    "!",
)


@dataclass
class CondaEnvironment(BaseEnvironment[Path]):
    BACKEND_NAME: ClassVar[str] = "conda"

    packages: List[str] = field(default_factory=list)
    python_version: Optional[str] = None
    env_dict: Optional[Dict[str, Any]] = None

    @classmethod
    def from_config(
        cls,
        config: Dict[str, Any],
        settings: IsolateSettings = DEFAULT_SETTINGS,
    ) -> BaseEnvironment:
        if config.get('env_dict') and config.get('env_yml_str'):
            raise EnvironmentCreationError("Either env_dict or env_yml_str can be provided, not both!")
        if config.get('env_yml_str'):
            config['env_dict'] = yaml.safe_load(config['env_yml_str'])
            del config['env_yml_str']
        environment = cls(**config)
        environment.apply_settings(settings)
        return environment

    @property
    def key(self) -> str:
        if self.env_dict:
            return sha256_digest_of(str(self._compute_dependencies()))
        return sha256_digest_of(*self._compute_dependencies())

    def _compute_dependencies(self) -> List[Any]:
        if self.env_dict:
            user_dependencies = self.env_dict.get('dependencies', []).copy()
        else:
            user_dependencies = self.packages.copy()
        for raw_requirement in user_dependencies:
            # It could be 'pip': [...]
            if type(raw_requirement) is dict:
                continue
            # Get rid of all whitespace characters (python = 3.8 becomes python=3.8)
            raw_requirement = raw_requirement.replace(" ", "")
            if not raw_requirement.startswith("python"):
                continue

            # Ensure that the package is either python or python followed
            # by a version specifier. Examples:
            #  - python # OK
            #  - python=3.8 # OK
            #  - python>=3.8 # OK
            #  - python-user-toolkit # NOT OK
            #  - pythonhelp!=1.0 # NOT OK

            python_suffix = raw_requirement[len("python") :]
            if (
                python_suffix
                and python_suffix[0] not in _CONDA_VERSION_IDENTIFIER_CHARS
            ):
                continue

            raise EnvironmentCreationError(
                "Python version can not be specified by packages (it needs to be passed as `python_version` option)"
            )

        # Now that we verified that the user did not specify the Python version
        # we can add it by ourselves
        target_python = self.python_version or active_python()
        user_dependencies.append(f"python={target_python}")
        return user_dependencies

    def create(self) -> Path:
        env_path = self.settings.cache_dir_for(self)
        with self.settings.cache_lock_for(env_path):
            if env_path.exists():
                return env_path

            if self.env_dict:
                self.env_dict['dependencies'] = self._compute_dependencies()
                with tempfile.NamedTemporaryFile(mode='w', suffix='.yml') as tf:
                    yaml.dump(self.env_dict, tf)
                    tf.flush()
                    try:
                        self._run_conda(
                            "env",
                            "create",
                            "-f",
                            tf.name,
                            "--prefix",
                            env_path
                        )
                    except subprocess.SubprocessError as exc:
                        raise EnvironmentCreationError("Failure during 'conda create'") from exc

            else:
                # Since our agent needs Python to be installed (at very least)
                # we need to make sure that the base environment is created with
                # the same Python version as the one that is used to run the
                # isolate agent.
                dependencies = self._compute_dependencies()

                self.log(f"Creating the environment at '{env_path}'")
                self.log(f"Installing packages: {', '.join(dependencies)}")

                try:
                    self._run_conda(
                        "create",
                        "--yes",
                        "--prefix",
                        env_path,
                        *dependencies,
                    )
                except subprocess.SubprocessError as exc:
                    raise EnvironmentCreationError("Failure during 'conda create'") from exc

            self.log(f"New environment cached at '{env_path}'")
            return env_path

    def destroy(self, connection_key: Path) -> None:
        with self.settings.cache_lock_for(connection_key):
            # It might be destroyed already (when we are awaiting
            # for the lock to be released).
            if not connection_key.exists():
                return

            self._run_conda(
                "remove",
                "--yes",
                "--all",
                "--prefix",
                connection_key,
            )

    def _run_conda(self, *args: Any) -> None:
        conda_executable = _get_conda_executable()
        with logged_io(self.log) as (stdout, stderr):
            subprocess.check_call(
                [conda_executable, *args],
                stdout=stdout,
                stderr=stderr,
            )

    def exists(self) -> bool:
        path = self.settings.cache_dir_for(self)
        return path.exists()

    def open_connection(self, connection_key: Path) -> PythonIPC:
        return PythonIPC(self, connection_key)


@functools.lru_cache(1)
def _get_conda_executable() -> Path:
    for path in [_ISOLATE_CONDA_HOME, None]:
        conda_path = shutil.which(_CONDA_COMMAND, path=path)
        if conda_path is not None:
            return Path(conda_path)
    else:
        # TODO: we should probably show some instructions on how you
        # can install conda here.
        raise FileNotFoundError(
            "Could not find conda executable. If conda executable is not available by default, please point isolate "
            " to the path where conda binary is available 'ISOLATE_CONDA_HOME'."
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Type, Union

import importlib_metadata

if TYPE_CHECKING:
    from isolate.backends import BaseEnvironment

# Any new environments can register themselves during package installation
# time by simply adding an entry point to the `isolate.environment` group.
_ENTRY_POINT = "isolate.backends"

_ENTRY_POINTS: Dict[str, importlib_metadata.EntryPoint] = {}
_ENVIRONMENTS: Dict[str, Type["BaseEnvironment"]] = {}


def _reload_registry() -> None:
    entry_points = importlib_metadata.entry_points()
    _ENTRY_POINTS.update(
        {
            # We are not immediately loading the backend class here
            # since it might cause importing modules that we won't be
            # using at all.
            entry_point.name: entry_point
            for entry_point in entry_points.select(group=_ENTRY_POINT)
        }
    )


_reload_registry()


def prepare_environment(
    kind: str,
    **kwargs: Any,
) -> BaseEnvironment:
    """Get the environment for the given `kind` with the given `config`."""
    from isolate.backends.settings import DEFAULT_SETTINGS

    if kind not in _ENVIRONMENTS:
        entry_point = _ENTRY_POINTS.get(kind)
        if not entry_point:
            raise ValueError(f"Unknown environment: '{kind}'")

        _ENVIRONMENTS[kind] = entry_point.load()

    settings = kwargs.pop("context", DEFAULT_SETTINGS)
    return _ENVIRONMENTS[kind].from_config(config=kwargs, settings=settings)



from google.protobuf.json_format import MessageToDict as struct_to_dict
from google.protobuf.struct_pb2 import Struct

# Inherit everything from the gRPC connection handler.
from isolate.connections.grpc.definitions import *
from isolate.server.definitions.server_pb2 import *
from isolate.server.definitions.server_pb2_grpc import (
    IsolateServicer,
    IsolateStub,
)
from isolate.server.definitions.server_pb2_grpc import (
    add_IsolateServicer_to_server as register_isolate,
)



"""A common gRPC interface for both the gRPC connection implementation
and the Isolate Server to share."""

import functools
from typing import TYPE_CHECKING, Any, Optional

from isolate.connections.common import load_serialized_object, serialize_object
from isolate.connections.grpc import definitions
from isolate.logs import Log, LogLevel, LogSource


@functools.singledispatch
def from_grpc(message: definitions.Message) -> Any:
    """Materialize a gRPC message into a Python object."""
    wrong_type = type(message).__name__
    raise NotImplementedError(f"Can't convert {wrong_type} to a Python object.")


@functools.singledispatch
def to_grpc(obj: Any) -> definitions.Message:
    """Convert a Python object into a gRPC message."""
    wrong_type = type(obj).__name__
    raise NotImplementedError(f"Cannot convert {wrong_type} to a gRPC message.")


@from_grpc.register
def _(message: definitions.SerializedObject) -> Any:
    return load_serialized_object(
        message.method,
        message.definition,
        was_it_raised=message.was_it_raised,
        stringized_traceback=message.stringized_traceback,
    )


@from_grpc.register
def _(message: definitions.Log) -> Log:
    source = LogSource(definitions.LogSource.Name(message.source).lower())
    level = LogLevel(definitions.LogLevel.Name(message.level).lower())
    return Log(
        message=message.message,
        source=source,
        level=level,
    )


@to_grpc.register
def _(obj: Log) -> definitions.Log:
    return definitions.Log(
        message=obj.message,
        source=definitions.LogSource.Value(obj.source.name.upper()),
        level=definitions.LogLevel.Value(obj.level.name.upper()),
    )


def to_serialized_object(
    obj: Any,
    method: str,
    was_it_raised: bool = False,
    stringized_traceback: Optional[str] = None,
) -> definitions.SerializedObject:
    """Convert a Python object into a gRPC message."""
    return definitions.SerializedObject(
        method=method,
        definition=serialize_object(method, obj),
        was_it_raised=was_it_raised,
        stringized_traceback=stringized_traceback,
    )



from __future__ import annotations

import functools
import os
import shutil
import subprocess
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, ClassVar, Dict, List, Optional

from isolate.backends import BaseEnvironment, EnvironmentCreationError
from isolate.backends.common import active_python, logged_io, sha256_digest_of
from isolate.backends.settings import DEFAULT_SETTINGS, IsolateSettings
from isolate.connections import PythonIPC

_PYENV_EXECUTABLE_NAME = "pyenv"
_PYENV_EXECUTABLE_PATH = os.environ.get("ISOLATE_PYENV_EXECUTABLE")


@dataclass
class PyenvEnvironment(BaseEnvironment[Path]):
    BACKEND_NAME: ClassVar[str] = "pyenv"

    python_version: str

    @classmethod
    def from_config(
        cls,
        config: Dict[str, Any],
        settings: IsolateSettings = DEFAULT_SETTINGS,
    ) -> BaseEnvironment:
        environment = cls(**config)
        environment.apply_settings(settings)
        return environment

    @property
    def key(self) -> str:
        return os.path.join("versions", self.python_version)

    def create(self) -> Path:
        pyenv = _get_pyenv_executable()
        env_path = self.settings.cache_dir_for(self)
        with self.settings.cache_lock_for(env_path):
            # PyEnv installs* the Python versions under $root/versions/$version, where we
            # use versions/$version as the key and $root as the base directory (for pyenv).
            #
            # [0]: https://github.com/pyenv/pyenv#locating-pyenv-provided-python-installations
            pyenv_root = env_path.parent.parent
            prefix = self._try_get_prefix(pyenv, pyenv_root)
            if prefix is None:
                self._install_python(pyenv, pyenv_root)
                prefix = self._try_get_prefix(pyenv, pyenv_root)
                if not prefix:
                    raise EnvironmentCreationError(
                        f"Python {self.python_version} must have been installed by now."
                    )

            assert prefix is not None
            return prefix

    def _try_get_prefix(self, pyenv: Path, root_path: Path) -> Optional[Path]:
        try:
            prefix = subprocess.check_output(
                [pyenv, "prefix", self.python_version],
                env={**os.environ, "PYENV_ROOT": str(root_path)},
                text=True,
                stderr=subprocess.PIPE,
            )
        except subprocess.CalledProcessError as exc:
            if "not installed" in exc.stderr:
                return None
            raise EnvironmentCreationError(
                f"Failed to get the prefix for Python {self.python_version}.\n"
                f"{exc.stdout}\n{exc.stderr}"
            )

        return Path(prefix.strip())

    def _install_python(self, pyenv: Path, root_path: Path) -> None:
        with logged_io(self.log) as (stdout, stderr):
            try:
                subprocess.check_call(
                    [pyenv, "install", "--skip-existing", self.python_version],
                    env={**os.environ, "PYENV_ROOT": str(root_path)},
                    stdout=stdout,
                    stderr=stderr,
                )
            except subprocess.CalledProcessError as exc:
                raise EnvironmentCreationError(
                    f"Failed to install Python {self.python_version} via pyenv.\n"
                )

    def destroy(self, connection_key: Path) -> None:
        pyenv = _get_pyenv_executable()
        with self.settings.cache_lock_for(connection_key):
            # It might be destroyed already (when we are awaiting
            # for the lock to be released).
            if not connection_key.exists():
                return None

            pyenv_root = connection_key.parent.parent
            with logged_io(self.log) as (stdout, stderr):
                subprocess.check_call(
                    [pyenv, "uninstall", "-f", connection_key.name],
                    env={**os.environ, "PYENV_ROOT": str(pyenv_root)},
                    stdout=stdout,
                    stderr=stderr,
                )

    def exists(self) -> bool:
        pyenv = _get_pyenv_executable()
        cache_dir = self.settings.cache_dir_for(self)
        with self.settings.cache_lock_for(cache_dir):
            pyenv_root = cache_dir.parent.parent
            prefix = self._try_get_prefix(pyenv, pyenv_root)
            return prefix is not None

    def open_connection(self, connection_key: Path) -> PythonIPC:
        return PythonIPC(self, connection_key)


@functools.lru_cache(1)
def _get_pyenv_executable() -> Path:
    if _PYENV_EXECUTABLE_PATH:
        if not os.path.exists(_PYENV_EXECUTABLE_PATH):
            raise EnvironmentCreationError(
                f"Path to pyenv executable not found! ISOLATE_PYENV_EXECUTABLE variable: {_PYENV_EXECUTABLE_PATH!r}"
            )
        return Path(_PYENV_EXECUTABLE_PATH)

    pyenv_path = shutil.which(_PYENV_EXECUTABLE_NAME)
    if pyenv_path is None:
        raise FileNotFoundError(
            "Could not find the pyenv executable. If pyenv is not already installed in your system, please"
            "install it first. If it is not in your PATH, then point ISOLATE_PYENV_COMMAND to the absolute path of the"
            "pyenv executable."
        )
    return Path(pyenv_path)



import os
import traceback
from concurrent import futures
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass, field, replace
from functools import partial
from queue import Empty as QueueEmpty
from queue import Queue
from typing import Callable, Iterator, cast

import grpc
from grpc import ServicerContext, StatusCode

from isolate.backends import EnvironmentCreationError, IsolateSettings
from isolate.backends.common import active_python
from isolate.backends.local import LocalPythonEnvironment
from isolate.backends.virtualenv import VirtualPythonEnvironment
from isolate.connections.grpc import AgentError, LocalPythonGRPC
from isolate.logs import Log, LogLevel, LogSource
from isolate.server import definitions
from isolate.server.interface import from_grpc, to_grpc

# Whether to inherit all the packages from the current environment or not.
INHERIT_FROM_LOCAL = os.getenv("ISOLATE_INHERIT_FROM_LOCAL") == "1"

# Number of threads that the gRPC server will use.
MAX_THREADS = int(os.getenv("MAX_THREADS", 5))
_AGENT_REQUIREMENTS_TXT = os.getenv("AGENT_REQUIREMENTS_TXT")

if _AGENT_REQUIREMENTS_TXT is not None:
    with open(_AGENT_REQUIREMENTS_TXT) as stream:
        AGENT_REQUIREMENTS = stream.read().splitlines()
else:
    AGENT_REQUIREMENTS = []


# Number of seconds to observe the queue before checking the termination
# event.
_Q_WAIT_DELAY = 0.1


@dataclass
class IsolateServicer(definitions.IsolateServicer):
    default_settings: IsolateSettings = field(default_factory=IsolateSettings)

    def Run(
        self,
        request: definitions.BoundFunction,
        context: ServicerContext,
    ) -> Iterator[definitions.PartialRunResult]:
        messages: Queue[definitions.PartialRunResult] = Queue()
        environments = []
        for env in request.environments:
            try:
                environments.append(from_grpc(env))
            except ValueError:
                return self.abort_with_msg(
                    f"Unknown environment kind: {env.kind}",
                    context,
                )
            except TypeError as exc:
                return self.abort_with_msg(
                    f"Invalid environment parameter: {str(exc)}.",
                    context,
                )

        if not environments:
            return self.abort_with_msg(
                f"At least one environment must be specified for a run!",
                context,
            )

        run_settings = replace(
            self.default_settings,
            log_hook=partial(_add_log_to_queue, messages),
            serialization_method=request.function.method,
        )

        for environment in environments:
            environment.apply_settings(run_settings)

        primary_environment = environments[0]

        if AGENT_REQUIREMENTS:
            python_version = getattr(
                primary_environment, "python_version", active_python()
            )
            agent_environ = VirtualPythonEnvironment(
                requirements=AGENT_REQUIREMENTS,
                python_version=python_version,
            )
            agent_environ.apply_settings(run_settings)
            environments.insert(1, agent_environ)

        extra_inheritance_paths = []
        if INHERIT_FROM_LOCAL:
            local_environment = LocalPythonEnvironment()
            extra_inheritance_paths.append(local_environment.create())

        with ThreadPoolExecutor(max_workers=1) as local_pool:
            environment_paths = []
            for environment in environments:
                future = local_pool.submit(environment.create)
                yield from self.watch_queue_until_completed(messages, future.done)
                try:
                    # Assuming that the iterator above only stops yielding once
                    # the future is completed, the timeout here should be redundant
                    # but it is just in case.
                    environment_paths.append(future.result(timeout=0.1))
                except EnvironmentCreationError:
                    return self.abort_with_msg(
                        f"A problem occurred while creating the environment.",
                        context,
                    )

            primary_path, *inheritance_paths = environment_paths
            inheritance_paths.extend(extra_inheritance_paths)
            primary_environment = environments[0]

            with LocalPythonGRPC(
                primary_environment,
                primary_path,
                extra_inheritance_paths=inheritance_paths,
            ) as connection:
                future = local_pool.submit(
                    _proxy_to_queue,
                    queue=messages,
                    connection=cast(LocalPythonGRPC, connection),
                    input=request.function,
                )

                # Unlike above; we are not interested in the result value of future
                # here, since it will be already transferred to other side without
                # us even seeing (through the queue).
                yield from self.watch_queue_until_completed(messages, future.done)

                # But we still have to check whether there were any errors raised
                # during the execution, and handle them accordingly.
                exception = future.exception(timeout=0.1)
                if exception is not None:
                    for line in traceback.format_exception(
                        type(exception), exception, exception.__traceback__
                    ):
                        yield from self.log(line, level=LogLevel.ERROR)
                    if isinstance(exception, AgentError):
                        return self.abort_with_msg(
                            str(exception),
                            context,
                            code=StatusCode.ABORTED,
                        )
                    else:
                        return self.abort_with_msg(
                            f"An unexpected error occurred.",
                            context,
                            code=StatusCode.UNKNOWN,
                        )

    def watch_queue_until_completed(
        self, queue: Queue, is_completed: Callable[[], bool]
    ) -> Iterator[definitions.PartialRunResult]:
        """Watch the given queue until the is_completed function returns True. Note that even
        if the function is completed, this function might not finish until the queue is empty."""
        while not is_completed():
            try:
                yield queue.get(timeout=_Q_WAIT_DELAY)
            except QueueEmpty:
                continue

        # Clear the final messages
        while not queue.empty():
            try:
                yield queue.get_nowait()
            except QueueEmpty:
                continue

    def log(
        self,
        message: str,
        level: LogLevel = LogLevel.TRACE,
        source: LogSource = LogSource.BRIDGE,
    ) -> Iterator[definitions.PartialRunResult]:
        log = to_grpc(Log(message, level=level, source=source))
        log = cast(definitions.Log, log)
        yield definitions.PartialRunResult(result=None, is_complete=False, logs=[log])

    def abort_with_msg(
        self,
        message: str,
        context: ServicerContext,
        *,
        code: StatusCode = StatusCode.INVALID_ARGUMENT,
    ) -> None:
        context.set_code(code)
        context.set_details(message)
        return None


def _proxy_to_queue(
    queue: Queue,
    connection: LocalPythonGRPC,
    input: definitions.SerializedObject,
) -> None:
    for message in connection._run_through_grpc(input):
        queue.put_nowait(message)


def _add_log_to_queue(messages: Queue, log: Log) -> None:
    grpc_log = cast(definitions.Log, to_grpc(log))
    grpc_result = definitions.PartialRunResult(
        is_complete=False,
        logs=[grpc_log],
        result=None,
    )
    messages.put_nowait(grpc_result)


def main() -> None:
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=MAX_THREADS))
    definitions.register_isolate(IsolateServicer(), server)

    server.add_insecure_port(f"[::]:50001")
    print("Started listening at localhost:50001")

    server.start()
    server.wait_for_termination()


if __name__ == "__main__":
    main()



from __future__ import annotations

import base64
import importlib
import subprocess
import time
from contextlib import ExitStack, closing
from dataclasses import dataclass
from multiprocessing.connection import Connection, Listener
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ContextManager,
    List,
    Tuple,
    Union,
)

from isolate.backends import (
    BasicCallable,
    CallResultType,
    EnvironmentConnection,
)
from isolate.connections._local import PythonExecutionBase, agent_startup
from isolate.connections.common import prepare_exc
from isolate.connections.ipc import agent
from isolate.logs import LogLevel, LogSource

if TYPE_CHECKING:
    # Somhow mypy can't figure out that `ConnectionWrapper`
    # really exists.
    class ConnectionWrapper(Connection):
        def __init__(
            self,
            connection: Any,
            loads: Callable[[bytes], Any],
            dumps: Callable[[Any], bytes],
        ) -> None:
            ...

        def recv(self) -> Any:
            ...

        def send(self, value: Any) -> None:
            ...

        def close(self) -> None:
            ...

else:
    from multiprocessing.connection import ConnectionWrapper


class AgentListener(Listener):
    """A custom listener that can use any available serialization method
    to communicate with the child process."""

    def __init__(self, backend_name: str, *args: Any, **kwargs: Any) -> None:
        self.serialization_backend = loadserialization_method(backend_name)
        super().__init__(*args, **kwargs)

    def accept(self) -> Connection:
        return ConnectionWrapper(
            super().accept(),
            dumps=self.serialization_backend.dumps,
            loads=self.serialization_backend.loads,
        )


def loadserialization_method(backend_name: str) -> Any:
    # TODO(feat): This should probably throw a better error if the
    # given backend does not exist.
    return importlib.import_module(backend_name)


def encode_service_address(address: Tuple[str, int]) -> str:
    host, port = address
    return base64.b64encode(f"{host}:{port}".encode()).decode("utf-8")


@dataclass
class IsolatedProcessConnection(EnvironmentConnection):
    """A generic IPC implementation for running the isolate backend
    in a separated process.

    Each implementation needs to define a start_process method to
    spawn the agent."""

    # The amount of seconds to wait before checking whether the
    # isolated process has exited or not.
    _DEFER_THRESHOLD = 0.25

    def start_process(
        self,
        connection: AgentListener,
        *args: Any,
        **kwargs: Any,
    ) -> ContextManager[subprocess.Popen]:
        """Start the agent process."""
        raise NotImplementedError

    def run(
        self,
        executable: BasicCallable,
        *args: Any,
        **kwargs: Any,
    ) -> CallResultType:
        """Spawn an agent process using the given environment, run the given
        `executable` in that process, and return the result object back."""

        with ExitStack() as stack:
            # IPC flow is the following:
            #  1. [controller]: Create the socket server
            #  2. [controller]: Spawn the call agent with the socket address
            #  3.      [agent]: Connect to the socket server
            #  4. [controller]: Accept the incoming connection request
            #  5. [controller]: Send the executable over the established bridge
            #  6.      [agent]: Receive the executable from the bridge
            #  7.      [agent]: Execute the executable and once done send the result back
            #  8. [controller]: Loop until either the isolated process exits or sends any
            #                   data (will be interpreted as a tuple of two mutually exclusive
            #                   objects, either a result object or an exception to be raised).
            #

            self.log("Starting the controller bridge.")
            controller_service = stack.enter_context(
                AgentListener(
                    self.environment.settings.serialization_method,
                    family="AF_INET",
                )
            )

            self.log(
                f"Controller server is listening at {controller_service.address}."
                " Attempting to start the agent process."
            )
            assert not (args or kwargs), "run() should not receive any arguments."
            isolated_process = stack.enter_context(
                self.start_process(controller_service, *args, **kwargs)
            )

            # TODO(fix): this might hang if the agent process crashes before it can
            # connect to the controller bridge.
            self.log(
                f"Awaiting agent process of {isolated_process.pid}"
                " to establish a connection."
            )
            established_connection = stack.enter_context(
                closing(controller_service.accept())
            )

            self.log("Bridge between controller and the agent has been established.")
            established_connection.send(executable)

            self.log("Executable has been sent, awaiting execution result.")
            return self.poll_until_result(
                isolated_process,
                established_connection,
            )

    def poll_until_result(
        self,
        process: subprocess.Popen,
        connection: Connection,
    ) -> CallResultType:
        """Take the given process, and poll until either it exits or returns
        a result object."""

        while not connection.poll():
            # Normally, if we do connection.read() without having this loop
            # it is going to block us indefinitely (even if the underlying
            # process has crashed). We can use a combination of process.poll
            # and connection.poll to check if the process is alive and has data
            # to move forward.
            if process.poll():
                break

            # For preventing busy waiting, we can sleep for a bit
            # and let other threads run.
            time.sleep(self._DEFER_THRESHOLD)
            continue

        if not connection.poll():
            # If the process has exited but there is still no data, we
            # can assume something terrible has happened.
            raise OSError(
                "The isolated process has exited unexpectedly with code "
                f"'{process.poll()}' without sending any data back."
            )

        # TODO(fix): handle EOFError that might happen here (e.g. problematic
        # serialization might cause it).
        result, did_it_raise, stringized_traceback = connection.recv()

        if did_it_raise:
            raise prepare_exc(result, stringized_traceback=stringized_traceback)
        else:
            assert stringized_traceback is None
            return result


@dataclass
class PythonIPC(PythonExecutionBase[AgentListener], IsolatedProcessConnection):
    def get_python_cmd(
        self,
        executable: Path,
        connection: AgentListener,
    ) -> List[Union[str, Path]]:
        assert isinstance(connection.address, tuple)
        return [
            executable,
            agent_startup.__file__,
            agent.__file__,
            encode_service_address(connection.address),
            # TODO(feat): we probably should check if the given backend is installed
            # on the remote interpreter, otherwise it will fail without establishing
            # the connection with the bridge.
            "--serialization-backend",
            self.environment.settings.serialization_method,
        ]

    def handle_agent_log(self, line: str, level: LogLevel) -> None:
        # TODO: we probably should create a new fd and pass it as
        # one of the the arguments to the child process. Then everything
        # from that fd can be automatically logged as originating from the
        # bridge.

        # Agent can produce [trace] messages, so change the log
        # level to it if this does not originate from the user.
        if line.startswith("[trace]"):
            line = line.replace("[trace]", "", 1)
            level = LogLevel.TRACE
            source = LogSource.BRIDGE
        else:
            source = LogSource.USER

        self.log(line, level=level, source=source)



# This file defines an "isolate" agent for inter-process communication over
# sockets. It is spawned by the controller process with a single argument (a
# base64 encoded server address) and expected to go through the following procedures:
#   1. Decode the given address
#   2. Create a connection to the transmission bridge using the address
#   3. Receive a callable object from the bridge
#   4. Execute the callable object
#   5. Send the result back to the bridge
#   6. Exit
#
# Up until to point 4, the agent process has no way of transmitting information
# to the controller so it should use the stderr/stdout channels appropriately. After
# the executable is received, the controller process will switch to the listening mode
# and wait for agent to return something. The expected object is a tuple of two objects
# one being the actual result of the given callable, and the other one is a boolean flag
# indicating whether the callable has raised an exception or not.

import base64
import importlib
import os
import sys
import time
import traceback
from argparse import ArgumentParser
from contextlib import closing
from multiprocessing.connection import Client
from typing import TYPE_CHECKING, Any, Callable, ContextManager, Tuple

if TYPE_CHECKING:
    # Somhow mypy can't figure out that `ConnectionWrapper`
    # really exists.
    class ConnectionWrapper:
        def __init__(
            self,
            connection: Any,
            loads: Callable[[bytes], Any],
            dumps: Callable[[Any], bytes],
        ) -> None:
            ...

        def recv(self) -> Any:
            ...

        def send(self, value: Any) -> None:
            ...

        def close(self) -> None:
            ...

else:
    from multiprocessing.connection import ConnectionWrapper


def decode_service_address(address: str) -> Tuple[str, int]:
    host, port = base64.b64decode(address).decode("utf-8").rsplit(":", 1)
    return host, int(port)


def child_connection(
    serialization_method: str, address: Tuple[str, int]
) -> ContextManager[ConnectionWrapper]:
    serialization_backend = importlib.import_module(serialization_method)
    return closing(
        ConnectionWrapper(
            Client(address),
            loads=serialization_backend.loads,
            dumps=serialization_backend.dumps,
        )
    )


IS_DEBUG_MODE = os.getenv("ISOLATE_ENABLE_DEBUGGING") == "1"
DEBUG_TIMEOUT = 60 * 15


def run_client(
    serialization_method: str, address: Tuple[str, int], *, with_pdb: bool = False
) -> None:
    # Debug Mode
    # ==========
    #
    # Isolated processes are really tricky to debug properly
    # so we want to have a smooth way into the process and see
    # what is really going on in the case of errors.
    #
    # For using the debug mode, you first need to set ISOLATE_ENABLE_DEBUGGING
    # environment variable to "1" from your controller process. This will
    # make the isolated process hang at the initialization, and make it print
    # the instructions to connect to the controller process.
    #
    # On a separate shell (while letting the the controller process hang), you can
    # execute the given command to drop into the PDB (Python Debugger). With that
    # you can observe each step of the connection and run process.

    if with_pdb:
        # This condition will only be activated if we want to
        # debug the isolated process by passing the --with-pdb
        # flag when executing the binary.
        import pdb

        pdb.set_trace()

    print(f"[trace] Trying to create a connection to {address}")
    # TODO(feat): this should probably run in a loop instead of
    # receiving a single function and then exitting immediately.
    with child_connection(serialization_method, address) as connection:
        print(f"[trace] Created child connection to {address}")
        callable = connection.recv()
        print(f"[trace] Received the callable at {address}")

        result = None
        did_it_raise = False
        stringized_tb = None
        try:
            result = callable()
        except BaseException as exc:
            result = exc
            did_it_raise = True
            num_frames = len(traceback.extract_stack()[:-4])
            stringized_tb = "".join(traceback.format_exc(limit=-num_frames))
        finally:
            try:
                connection.send((result, did_it_raise, stringized_tb))
            except BaseException:
                if did_it_raise:
                    # If we can't even send it through the connection
                    # still try to dump it to the stderr as the last
                    # resort.
                    assert isinstance(result, BaseException)
                    traceback.print_exception(
                        type(result),
                        result,
                        result.__traceback__,
                    )
                raise


def _get_shell_bootstrap() -> str:
    # Return a string that contains environment variables that
    # might be used during isolated hook's execution.
    return " ".join(
        f"{session_variable}={os.getenv(session_variable)}"
        for session_variable in [
            # PYTHONPATH is customized by the Extended Environment IPC
            # system to make sure that the isolated process can
            # import stuff from the primary environment. Without this
            # the isolated process will not be able to run properly
            # on the newly created debug session.
            "PYTHONPATH",
        ]
        if session_variable in os.environ
    )


def main() -> int:
    print(f"[trace] Starting the isolated process at PID {os.getpid()}")

    parser = ArgumentParser()
    parser.add_argument("listen_at")
    parser.add_argument("--with-pdb", action="store_true", default=False)
    parser.add_argument("--serialization-backend", default="pickle")

    options = parser.parse_args()
    if IS_DEBUG_MODE:
        assert not options.with_pdb, "--with-pdb can't be used in the debug mode"
        message = "=" * 60
        message += "\n" * 3
        message += "Debug mode successfully activated. You can start your debugging session with the following command:\n"
        message += (
            f"    $ {_get_shell_bootstrap()}\\\n     "
            f"{sys.executable} {os.path.abspath(__file__)} "
            f"--serialization-backend {options.serialization_backend} "
            f"--with-pdb {options.listen_at}"
        )
        message += "\n" * 3
        message += "=" * 60
        print(message)
        time.sleep(DEBUG_TIMEOUT)

    serialization_method = options.serialization_backend
    address = decode_service_address(options.listen_at)
    run_client(serialization_method, address, with_pdb=options.with_pdb)
    return 0


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: agent.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from isolate.connections.grpc.definitions import common_pb2 as common__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x0b\x61gent.proto\x1a\x0c\x63ommon.proto28\n\x05\x41gent\x12/\n\x03Run\x12\x11.SerializedObject\x1a\x11.PartialRunResult"\x00\x30\x01\x62\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "agent_pb2", globals())
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _AGENT._serialized_start = 29
    _AGENT._serialized_end = 85
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

import importlib
import os
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Iterator, Optional, cast

from tblib import Traceback, TracebackParseError

if TYPE_CHECKING:
    from typing import Protocol

    class SerializationBackend(Protocol):
        def loads(self, data: bytes) -> Any:
            ...

        def dumps(self, obj: Any) -> bytes:
            ...


AGENT_SIGNATURE = "IS_ISOLATE_AGENT"


class SerializationError(Exception):
    """An error that happened during the serialization process."""


@contextmanager
def _step(message: str) -> Iterator[None]:
    """A context manager to capture every expression
    underneath it and if any of them fails for any reason
    then it will raise a SerializationError with the
    given message."""

    try:
        yield
    except BaseException as exception:
        raise SerializationError("Error while " + message) from exception


def as_serialization_method(backend: Any) -> SerializationBackend:
    """Ensures that the given backend has loads/dumps methods, and returns
    it as is (also convinces type checkers that the given object satisfies
    the serialization protocol)."""

    if not hasattr(backend, "loads") or not hasattr(backend, "dumps"):
        raise TypeError(
            f"The given serialization backend ({backend.__name__}) does "
            "not have one of the required methods (loads/dumps)."
        )

    return cast("SerializationBackend", backend)


def load_serialized_object(
    serialization_method: str,
    raw_object: bytes,
    *,
    was_it_raised: bool = False,
    stringized_traceback: Optional[str] = None,
) -> Any:
    """Load the given serialized object using the given serialization method. If
    anything fails, then a SerializationError will be raised. If the was_it_raised
    flag is set to true, then the given object will be raised as an exception (instead
    of being returned)."""

    with _step(f"preparing the serialization backend ({serialization_method})"):
        serialization_backend = as_serialization_method(
            importlib.import_module(serialization_method)
        )

    with _step("deserializing the given object"):
        result = serialization_backend.loads(raw_object)

    if was_it_raised:
        raise prepare_exc(result, stringized_traceback=stringized_traceback)
    else:
        return result


def serialize_object(serialization_method: str, object: Any) -> bytes:
    """Serialize the given object using the given serialization method. If
    anything fails, then a SerializationError will be raised."""

    with _step(f"preparing the serialization backend ({serialization_method})"):
        serialization_backend = as_serialization_method(
            importlib.import_module(serialization_method)
        )

    with _step("serializing the given object"):
        return serialization_backend.dumps(object)


def is_agent() -> bool:
    """Returns true if the current process is an isolate agent."""
    return os.environ.get(AGENT_SIGNATURE) == "1"


def prepare_exc(
    exc: BaseException,
    *,
    stringized_traceback: Optional[str] = None,
) -> BaseException:
    if stringized_traceback:
        try:
            traceback = Traceback.from_string(stringized_traceback).as_traceback()
        except TracebackParseError:
            traceback = None
    else:
        traceback = None

    exc.__traceback__ = traceback
    return exc



from google.protobuf.message import Message

from isolate.connections.grpc.definitions.agent_pb2 import *
from isolate.connections.grpc.definitions.agent_pb2_grpc import (
    AgentServicer,
    AgentStub,
)
from isolate.connections.grpc.definitions.agent_pb2_grpc import (
    add_AgentServicer_to_server as register_agent,
)
from isolate.connections.grpc.definitions.common_pb2 import *



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: common.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x0c\x63ommon.proto"\x89\x01\n\x10SerializedObject\x12\x0e\n\x06method\x18\x01 \x01(\t\x12\x12\n\ndefinition\x18\x02 \x01(\x0c\x12\x15\n\rwas_it_raised\x18\x03 \x01(\x08\x12!\n\x14stringized_traceback\x18\x04 \x01(\tH\x00\x88\x01\x01\x42\x17\n\x15_stringized_traceback"n\n\x10PartialRunResult\x12\x13\n\x0bis_complete\x18\x01 \x01(\x08\x12\x12\n\x04logs\x18\x02 \x03(\x0b\x32\x04.Log\x12&\n\x06result\x18\x03 \x01(\x0b\x32\x11.SerializedObjectH\x00\x88\x01\x01\x42\t\n\x07_result"L\n\x03Log\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\x1a\n\x06source\x18\x02 \x01(\x0e\x32\n.LogSource\x12\x18\n\x05level\x18\x03 \x01(\x0e\x32\t.LogLevel*.\n\tLogSource\x12\x0b\n\x07\x42UILDER\x10\x00\x12\n\n\x06\x42RIDGE\x10\x01\x12\x08\n\x04USER\x10\x02*Z\n\x08LogLevel\x12\t\n\x05TRACE\x10\x00\x12\t\n\x05\x44\x45\x42UG\x10\x01\x12\x08\n\x04INFO\x10\x02\x12\x0b\n\x07WARNING\x10\x03\x12\t\n\x05\x45RROR\x10\x04\x12\n\n\x06STDOUT\x10\x05\x12\n\n\x06STDERR\x10\x06\x62\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "common_pb2", globals())
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _LOGSOURCE._serialized_start = 346
    _LOGSOURCE._serialized_end = 392
    _LOGLEVEL._serialized_start = 394
    _LOGLEVEL._serialized_end = 484
    _SERIALIZEDOBJECT._serialized_start = 17
    _SERIALIZEDOBJECT._serialized_end = 154
    _PARTIALRUNRESULT._serialized_start = 156
    _PARTIALRUNRESULT._serialized_end = 266
    _LOG._serialized_start = 268
    _LOG._serialized_end = 344
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

import json
from dataclasses import dataclass
from typing import Any, ClassVar, Dict, List, Optional

import grpc

from isolate.backends import (
    BaseEnvironment,
    BasicCallable,
    CallResultType,
    EnvironmentConnection,
)
from isolate.backends.common import sha256_digest_of
from isolate.backends.settings import DEFAULT_SETTINGS, IsolateSettings
from isolate.server import interface
from isolate.server.definitions import (
    BoundFunction,
    EnvironmentDefinition,
    IsolateStub,
)


@dataclass
class IsolateServer(BaseEnvironment[List[EnvironmentDefinition]]):
    BACKEND_NAME: ClassVar[str] = "isolate-server"

    host: str
    target_environments: List[Dict[str, Any]]

    @classmethod
    def from_config(
        cls,
        config: Dict[str, Any],
        settings: IsolateSettings = DEFAULT_SETTINGS,
    ) -> BaseEnvironment:
        environment = cls(**config)
        environment.apply_settings(settings)

        return environment

    @property
    def key(self) -> str:
        return sha256_digest_of(
            self.host,
            json.dumps(self.target_environments),
        )

    def create(self) -> List[EnvironmentDefinition]:
        envs = []
        for env in self.target_environments:
            if not env.get("kind") or not env.get("configuration"):
                raise RuntimeError(f"`kind` or `configuration` key missing in: {env}")
            envs.append(
                EnvironmentDefinition(
                    kind=env["kind"],
                    configuration=interface.to_struct(env["configuration"]),
                )
            )
        return envs

    def exists(self) -> bool:
        return False

    def open_connection(
        self,
        connection_key: List[EnvironmentDefinition],
    ) -> IsolateServerConnection:
        return IsolateServerConnection(self, self.host, connection_key)


@dataclass
class IsolateServerConnection(EnvironmentConnection):
    host: str
    definitions: List[EnvironmentDefinition]
    _channel: Optional[grpc.Channel] = None

    def _acquire_channel(self) -> None:
        self._channel = grpc.insecure_channel(self.host)

    def _release_channel(self) -> None:
        if self._channel:
            self._channel.close()
            self._channel = None

    def __exit__(self, *args: Any) -> None:
        self._release_channel()

    def run(
        self,
        executable: BasicCallable,
        *args: Any,
        **kwargs: Any,
    ) -> CallResultType:
        if self._channel is None:
            self._acquire_channel()

        stub = IsolateStub(self._channel)
        request = BoundFunction(
            function=interface.to_serialized_object(
                executable,
                method=self.environment.settings.serialization_method,
                was_it_raised=False,
            ),
            environments=self.definitions,
        )

        return_value = []
        for result in stub.Run(request):
            for raw_log in result.logs:
                log = interface.from_grpc(raw_log)
                self.log(log.message, level=log.level, source=log.source)

            if result.is_complete:
                return_value.append(interface.from_grpc(result.result))

        if len(return_value) == 0:
            raise RuntimeError(
                "No result object was received from the server"
                " (it never set is_complete to True)."
            )
        elif len(return_value) > 1:
            raise RuntimeError(
                "Multiple result objects were received from the server"
                " (it set is_complete to True multiple times)."
            )
        else:
            return return_value[0]



from __future__ import annotations

import os
import shutil
import subprocess
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, ClassVar, Dict, List, Optional, Union

from isolate.backends import BaseEnvironment, EnvironmentCreationError
from isolate.backends.common import (
    active_python,
    get_executable_path,
    logged_io,
    sha256_digest_of,
)
from isolate.backends.settings import DEFAULT_SETTINGS, IsolateSettings
from isolate.connections import PythonIPC


@dataclass
class VirtualPythonEnvironment(BaseEnvironment[Path]):
    BACKEND_NAME: ClassVar[str] = "virtualenv"

    requirements: List[str] = field(default_factory=list)
    constraints_file: Optional[os.PathLike] = None
    python_version: Optional[str] = None
    extra_index_urls: List[str] = field(default_factory=list)

    @classmethod
    def from_config(
        cls,
        config: Dict[str, Any],
        settings: IsolateSettings = DEFAULT_SETTINGS,
    ) -> BaseEnvironment:
        environment = cls(**config)
        environment.apply_settings(settings)
        return environment

    @property
    def key(self) -> str:
        if self.constraints_file is not None:
            with open(self.constraints_file) as stream:
                constraints = stream.read().splitlines()
        else:
            constraints = []

        active_python_version = self.python_version or active_python()
        return sha256_digest_of(
            active_python_version,
            *self.requirements,
            *constraints,
            *self.extra_index_urls,
        )

    def install_requirements(self, path: Path) -> None:
        """Install the requirements of this environment using 'pip' to the
        given virtualenv path.

        If there are any constraint files specified, they will be also passed to
        the package resolver.
        """
        if not self.requirements:
            return None

        self.log(f"Installing requirements: {', '.join(self.requirements)}")

        pip_cmd: List[Union[str, os.PathLike]] = [
            get_executable_path(path, "pip"),
            "install",
            *self.requirements,
        ]
        if self.constraints_file:
            pip_cmd.extend(["-c", self.constraints_file])

        for extra_index_url in self.extra_index_urls:
            pip_cmd.extend(["--extra-index-url", extra_index_url])

        with logged_io(self.log) as (stdout, stderr):
            try:
                subprocess.check_call(
                    pip_cmd,
                    stdout=stdout,
                    stderr=stderr,
                )
            except subprocess.SubprocessError as exc:
                raise EnvironmentCreationError("Failure during 'pip install'.") from exc

    def _install_python_through_pyenv(self) -> str:
        from isolate.backends.pyenv import PyenvEnvironment

        self.log(
            f"Requested Python version of {self.python_version} is not available "
            "in the system, attempting to install it through pyenv."
        )

        pyenv = PyenvEnvironment.from_config(
            {"python_version": self.python_version},
            settings=self.settings,
        )
        return str(get_executable_path(pyenv.create(), "python"))

    def _decide_python(self) -> str:
        from virtualenv.discovery import builtin

        from isolate.backends.pyenv import _get_pyenv_executable

        interpreter = builtin.get_interpreter(self.python_version, ())
        if interpreter is not None:
            return interpreter.executable

        try:
            _get_pyenv_executable()
        except Exception:
            raise EnvironmentCreationError(
                f"Python {self.python_version} is not available in your "
                "system. Please install it first."
            ) from None
        else:
            return self._install_python_through_pyenv()

    def create(self) -> Path:
        from virtualenv import cli_run

        venv_path = self.settings.cache_dir_for(self)
        with self.settings.cache_lock_for(venv_path):
            if venv_path.exists():
                return venv_path

            self.log(f"Creating the environment at '{venv_path}'")

            args = [str(venv_path)]
            if self.python_version:
                args.append(f"--python={self._decide_python()}")

            try:
                cli_run(args)
            except (RuntimeError, OSError) as exc:
                raise EnvironmentCreationError(
                    f"Failed to create the environment at '{venv_path}'"
                ) from exc

            self.install_requirements(venv_path)

        self.log(f"New environment cached at '{venv_path}'")
        return venv_path

    def destroy(self, connection_key: Path) -> None:
        with self.settings.cache_lock_for(connection_key):
            # It might be destroyed already (when we are awaiting
            # for the lock to be released).
            if not connection_key.exists():
                return

            shutil.rmtree(connection_key)

    def exists(self) -> bool:
        path = self.settings.cache_dir_for(self)
        return path.exists()

    def open_connection(self, connection_key: Path) -> PythonIPC:
        return PythonIPC(self, connection_key)



from typing import Any, Dict

from isolate.backends import BaseEnvironment
from isolate.connections.grpc.interface import (
    from_grpc,
    to_grpc,
    to_serialized_object,
)
from isolate.server import definitions

__all__ = ["from_grpc", "to_grpc", "to_serialized_object", "to_struct"]


@from_grpc.register
def _(message: definitions.EnvironmentDefinition) -> BaseEnvironment:
    from isolate import prepare_environment

    return prepare_environment(
        message.kind,
        **definitions.struct_to_dict(message.configuration),
    )


def to_struct(data: Dict[str, Any]) -> definitions.Struct:
    struct = definitions.Struct()
    struct.update(data)
    return struct



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _LogSource:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _LogSourceEnumTypeWrapper(
    google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_LogSource.ValueType],
    builtins.type,
):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    BUILDER: _LogSource.ValueType  # 0
    BRIDGE: _LogSource.ValueType  # 1
    USER: _LogSource.ValueType  # 2

class LogSource(_LogSource, metaclass=_LogSourceEnumTypeWrapper): ...

BUILDER: LogSource.ValueType  # 0
BRIDGE: LogSource.ValueType  # 1
USER: LogSource.ValueType  # 2
global___LogSource = LogSource

class _LogLevel:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _LogLevelEnumTypeWrapper(
    google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_LogLevel.ValueType],
    builtins.type,
):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    TRACE: _LogLevel.ValueType  # 0
    DEBUG: _LogLevel.ValueType  # 1
    INFO: _LogLevel.ValueType  # 2
    WARNING: _LogLevel.ValueType  # 3
    ERROR: _LogLevel.ValueType  # 4
    STDOUT: _LogLevel.ValueType  # 5
    STDERR: _LogLevel.ValueType  # 6

class LogLevel(_LogLevel, metaclass=_LogLevelEnumTypeWrapper): ...

TRACE: LogLevel.ValueType  # 0
DEBUG: LogLevel.ValueType  # 1
INFO: LogLevel.ValueType  # 2
WARNING: LogLevel.ValueType  # 3
ERROR: LogLevel.ValueType  # 4
STDOUT: LogLevel.ValueType  # 5
STDERR: LogLevel.ValueType  # 6
global___LogLevel = LogLevel

@typing_extensions.final
class SerializedObject(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    METHOD_FIELD_NUMBER: builtins.int
    DEFINITION_FIELD_NUMBER: builtins.int
    WAS_IT_RAISED_FIELD_NUMBER: builtins.int
    STRINGIZED_TRACEBACK_FIELD_NUMBER: builtins.int
    method: builtins.str
    """The serialization method used to serialize the the raw_object. Must be
    present in the environment that is running the agent itself.
    """
    definition: builtins.bytes
    """The Python object serialized with the method above."""
    was_it_raised: builtins.bool
    """A flag indicating whether the given object was raised (e.g. an exception
    that was captured) or not.
    """
    stringized_traceback: builtins.str
    """The stringized version of the traceback, if it was raised."""
    def __init__(
        self,
        *,
        method: builtins.str = ...,
        definition: builtins.bytes = ...,
        was_it_raised: builtins.bool = ...,
        stringized_traceback: builtins.str | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "_stringized_traceback",
            b"_stringized_traceback",
            "stringized_traceback",
            b"stringized_traceback",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "_stringized_traceback",
            b"_stringized_traceback",
            "definition",
            b"definition",
            "method",
            b"method",
            "stringized_traceback",
            b"stringized_traceback",
            "was_it_raised",
            b"was_it_raised",
        ],
    ) -> None: ...
    def WhichOneof(
        self,
        oneof_group: typing_extensions.Literal[
            "_stringized_traceback", b"_stringized_traceback"
        ],
    ) -> typing_extensions.Literal["stringized_traceback"] | None: ...

global___SerializedObject = SerializedObject

@typing_extensions.final
class PartialRunResult(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    IS_COMPLETE_FIELD_NUMBER: builtins.int
    LOGS_FIELD_NUMBER: builtins.int
    RESULT_FIELD_NUMBER: builtins.int
    is_complete: builtins.bool
    """A flag indicating whether the run has completed."""
    @property
    def logs(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
        global___Log
    ]:
        """A list of logs collected during this partial execution. It does
        not include old logs.
        """
    @property
    def result(self) -> global___SerializedObject:
        """The result of the run, if it is complete."""
    def __init__(
        self,
        *,
        is_complete: builtins.bool = ...,
        logs: collections.abc.Iterable[global___Log] | None = ...,
        result: global___SerializedObject | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "_result", b"_result", "result", b"result"
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "_result",
            b"_result",
            "is_complete",
            b"is_complete",
            "logs",
            b"logs",
            "result",
            b"result",
        ],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["_result", b"_result"]
    ) -> typing_extensions.Literal["result"] | None: ...

global___PartialRunResult = PartialRunResult

@typing_extensions.final
class Log(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MESSAGE_FIELD_NUMBER: builtins.int
    SOURCE_FIELD_NUMBER: builtins.int
    LEVEL_FIELD_NUMBER: builtins.int
    message: builtins.str
    source: global___LogSource.ValueType
    level: global___LogLevel.ValueType
    def __init__(
        self,
        *,
        message: builtins.str = ...,
        source: global___LogSource.ValueType = ...,
        level: global___LogLevel.ValueType = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "level", b"level", "message", b"message", "source", b"source"
        ],
    ) -> None: ...

global___Log = Log



import importlib
from typing import TYPE_CHECKING

from isolate.connections.ipc import IsolatedProcessConnection, PythonIPC

if TYPE_CHECKING:
    from isolate.connections.grpc import LocalPythonGRPC


def __getattr__(name):
    if name == "LocalPythonGRPC":
        extra = "grpc"
        module_name = "isolate.connections.grpc"
    else:
        raise AttributeError(f"module {__name__!r} has no attribute {name!r}")

    try:
        module = importlib.import_module(module_name)
    except ImportError:
        raise AttributeError(
            f"For using {name!r} you need to install isolate with {extra!r} support."
            f'\n    $ pip install "isolate[{extra}]"'
        )

    return getattr(module, name)



from __future__ import annotations

import hashlib
import os
import shutil
import sysconfig
import threading
import time
from contextlib import contextmanager, suppress
from functools import lru_cache
from pathlib import Path
from typing import Any, Callable, Iterator, Optional, Tuple

# For ensuring that the lock is created and not forgotten
# (e.g. the process which acquires it crashes, so it is never
# released), we are going to check the lock file's mtime every
# _REVOKE_LOCK_DELAY seconds. If the mtime is older than that
# value, we are going to assume the lock is stale and revoke it.
_REVOKE_LOCK_DELAY = 30


@contextmanager
def lock_build_path(path: Path, lock_dir: Path) -> Iterator[None]:
    """Try to acquire a lock for all operations on the given 'path'. This guarantees
    that the path will not be modified by any other process while the lock is held."""
    lock_file = (lock_dir / path.name).with_suffix(".lock")
    while not _try_acquire(lock_file):
        time.sleep(0.05)
        continue

    with _keep_lock_alive(lock_file):
        yield


@contextmanager
def _keep_lock_alive(lock_file: Path) -> Iterator[None]:
    """Keep the lock file alive by updating its mtime as long
    as we are doing something in the cache."""
    event = threading.Event()

    def _keep_alive(per_beat_delay: float = 1) -> None:
        while not event.wait(per_beat_delay):
            lock_file.touch()
        lock_file.unlink()

    thread = threading.Thread(target=_keep_alive)
    try:
        thread.start()
        yield
    finally:
        event.set()
        thread.join()


def _try_acquire(lock_file: Path) -> bool:
    with suppress(FileNotFoundError):
        mtime = lock_file.stat().st_mtime
        if time.time() - mtime > _REVOKE_LOCK_DELAY:
            # The lock file exists, but it may be stale. Check the
            # mtime and if it is too old, revoke it.
            lock_file.unlink()

    try:
        lock_file.touch(exist_ok=False)
    except FileExistsError:
        return False
    else:
        return True


def get_executable_path(search_path: Path, executable_name: str) -> Path:
    """Return the path for the executable named 'executable_name' under
    the '/bin' directory of 'search_path'."""

    bin_dir = (search_path / "bin").as_posix()
    executable_path = shutil.which(executable_name, path=bin_dir)
    if executable_path is None:
        raise FileNotFoundError(
            f"Could not find '{executable_name}' in '{search_path}'. "
            f"Is the virtual environment corrupted?"
        )

    return Path(executable_path)


_MESSAGE_STREAM_DELAY = 0.1


def _observe_reader(
    reading_fd: int,
    termination_event: threading.Event,
    hook: Callable[[str], None],
) -> threading.Thread:
    """Starts a new thread that reads from the specified file descriptor
    ('reading_fd') and calls the 'hook' for each line until the EOF is
    reached.

    Caller is responsible for joining the thread.
    """

    assert not os.get_blocking(reading_fd), "reading_fd must be non-blocking"

    def _reader():
        with open(reading_fd) as stream:
            while not termination_event.wait(_MESSAGE_STREAM_DELAY):
                line = stream.readline()
                if not line:
                    continue

                # TODO: probably only strip the last newline, or
                # do not strip anything at all.
                hook(line.rstrip())

            # Once the termination is requested, read everything
            # that is left in the stream.
            for line in stream.readlines():
                hook(line.rstrip())

    observer_thread = threading.Thread(target=_reader)
    observer_thread.start()
    return observer_thread


def _unblocked_pipe() -> Tuple[int, int]:
    """Create a pair of unblocked pipes. This is actually
    the same as os.pipe2(os.O_NONBLOCK), but that is not
    available in MacOS so we have to do it manually."""

    read_fd, write_fd = os.pipe()
    os.set_blocking(read_fd, False)
    os.set_blocking(write_fd, False)
    return read_fd, write_fd


@contextmanager
def logged_io(
    stdout_hook: Callable[[str], None],
    stderr_hook: Optional[Callable[[str], None]] = None,
) -> Iterator[Tuple[int, int]]:
    """Open two new streams (for stdout and stderr, respectively) and start relaying all
    the output from them to the given hooks."""

    termination_event = threading.Event()

    stdout_reader_fd, stdout_writer_fd = _unblocked_pipe()
    stderr_reader_fd, stderr_writer_fd = _unblocked_pipe()

    stdout_observer = _observe_reader(
        stdout_reader_fd,
        termination_event,
        hook=stdout_hook,
    )
    stderr_observer = _observe_reader(
        stderr_reader_fd,
        termination_event,
        hook=stderr_hook or stdout_hook,
    )
    try:
        yield stdout_writer_fd, stderr_writer_fd
    finally:
        termination_event.set()
        try:
            stdout_observer.join(timeout=_MESSAGE_STREAM_DELAY)
            stderr_observer.join(timeout=_MESSAGE_STREAM_DELAY)
        except TimeoutError:
            raise RuntimeError("Log observers did not terminate in time.")


@lru_cache(maxsize=None)
def sha256_digest_of(*unique_fields: str, _join_char: str = "\n") -> str:
    """Return the SHA256 digest that corresponds to the combined version
    of 'unique_fields. The order is preserved."""

    inner_text = _join_char.join(unique_fields).encode()
    return hashlib.sha256(inner_text).hexdigest()


def active_python() -> str:
    """Return the active Python version that can be used for caching
    and re-creating this environment. Currently only covers major and
    minor versions (like 3.9); patch versions are ignored (like 3.9.4)."""
    return sysconfig.get_python_version()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: server.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2

from isolate.connections.grpc.definitions import common_pb2 as common__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x0cserver.proto\x1a\x0c\x63ommon.proto\x1a\x1cgoogle/protobuf/struct.proto"b\n\rBoundFunction\x12,\n\x0c\x65nvironments\x18\x01 \x03(\x0b\x32\x16.EnvironmentDefinition\x12#\n\x08\x66unction\x18\x02 \x01(\x0b\x32\x11.SerializedObject"U\n\x15\x45nvironmentDefinition\x12\x0c\n\x04kind\x18\x01 \x01(\t\x12.\n\rconfiguration\x18\x02 \x01(\x0b\x32\x17.google.protobuf.Struct27\n\x07Isolate\x12,\n\x03Run\x12\x0e.BoundFunction\x1a\x11.PartialRunResult"\x00\x30\x01\x62\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "server_pb2", globals())
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _BOUNDFUNCTION._serialized_start = 60
    _BOUNDFUNCTION._serialized_end = 158
    _ENVIRONMENTDEFINITION._serialized_start = 160
    _ENVIRONMENTDEFINITION._serialized_end = 245
    _ISOLATE._serialized_start = 247
    _ISOLATE._serialized_end = 302
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

import os
import subprocess
import sysconfig
from contextlib import contextmanager
from dataclasses import dataclass, field
from functools import partial
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generic,
    Iterator,
    List,
    TypeVar,
    Union,
)

from isolate.backends.common import get_executable_path, logged_io
from isolate.connections.common import AGENT_SIGNATURE
from isolate.logs import LogLevel

if TYPE_CHECKING:
    from isolate.backends import BaseEnvironment

ConnectionType = TypeVar("ConnectionType")


def binary_path_for(*search_paths: Path) -> str:
    """Return the binary search path for the given 'search_paths'.
    It will be a combination of the 'bin/' folders in them and
    the existing PATH environment variable."""

    paths = []
    for search_path in search_paths:
        path = sysconfig.get_path("scripts", vars={"base": search_path})
        paths.append(path)
        # Some distributions (conda) might include both a 'bin' and
        # a 'scripts' folder.

        auxilary_binary_path = search_path / "bin"
        if path != auxilary_binary_path and auxilary_binary_path.exists():
            paths.append(str(auxilary_binary_path))

    if "PATH" in os.environ:
        paths.append(os.environ["PATH"])

    return os.pathsep.join(paths)


def python_path_for(*search_paths: Path) -> str:
    """Return the PYTHONPATH for the library paths residing
    in the given 'search_paths'. The order of the paths is
    preserved."""
    assert len(search_paths) >= 1
    lib_paths = []
    for search_path in search_paths:
        # sysconfig defines the schema of the directories under
        # any comforming Python installation (like venv, conda, etc.).
        #
        # Be aware that Debian's system installation does not
        # comform sysconfig.
        raw_glob_expr = sysconfig.get_path(
            "purelib",
            vars={
                "base": search_path,
                "python_version": "*",
                "py_version_short": "*",
                "py_version_nodot": "*",
            },
        )
        relative_glob_expr = Path(raw_glob_expr).relative_to(search_path).as_posix()

        # Try to find expand the Python version in the path. This is
        # necessary for supporting multiple Python versions in the same
        # environment.
        for file in search_path.glob(relative_glob_expr):
            lib_paths.append(str(file))

    return os.pathsep.join(lib_paths)


@dataclass
class PythonExecutionBase(Generic[ConnectionType]):
    """A generic Python execution implementation that can trigger a new process
    and start watching stdout/stderr for the logs. The environment_path must be
    the base directory of a new Python environment (structure that complies with
    sysconfig). Python binary inside that environment will be used to run the
    agent process.

    If set, extra_inheritance_paths allows extending the custom package search
    system with additional environments. As an example, the current environment_path
    might point to an environment with numpy and the extra_inheritance_paths might
    point to an environment with pandas. In this case, the agent process will have
    access to both numpy and pandas. The order is important here, as the first
    path in the list will be the first one to be looked up (so if there is multiple
    versions of the same package in different environments, the one in the first
    path will take precedence). Dependency resolution and compatibility must be
    handled by the user."""

    environment: BaseEnvironment
    environment_path: Path
    extra_inheritance_paths: List[Path] = field(default_factory=list)

    @contextmanager
    def start_process(
        self,
        connection: ConnectionType,
        *args: Any,
        **kwargs: Any,
    ) -> Iterator[subprocess.Popen]:
        """Start the agent process with the Python binary available inside the
        bound environment."""

        python_executable = get_executable_path(self.environment_path, "python")
        with logged_io(
            partial(self.handle_agent_log, level=LogLevel.STDOUT),
            partial(self.handle_agent_log, level=LogLevel.STDERR),
        ) as (stdout, stderr):
            yield subprocess.Popen(
                self.get_python_cmd(python_executable, connection),
                env=self.get_env_vars(),
                stdout=stdout,
                stderr=stderr,
                text=True,
            )

    def get_env_vars(self) -> Dict[str, str]:
        """Return the environment variables to run the agent process with. By default
        PYTHONUNBUFFERED is set to 1 to ensure the prints to stdout/stderr are reflect
        immediately (so that we can seamlessly transfer logs)."""

        custom_vars = {}
        custom_vars[AGENT_SIGNATURE] = "1"
        custom_vars["PYTHONUNBUFFERED"] = "1"

        # NOTE: we don't have to manually set PYTHONPATH here if we are
        # using a single environment since python will automatically
        # use the proper path.
        if self.extra_inheritance_paths:
            # The order here should reflect the order of the inheritance
            # where the actual environment already takes precedence.
            custom_vars["PYTHONPATH"] = python_path_for(
                self.environment_path, *self.extra_inheritance_paths
            )

        # But the PATH must be always set since it will be not be
        # automatically set by Python (think of this as ./venv/bin/activate)
        custom_vars["PATH"] = binary_path_for(
            self.environment_path, *self.extra_inheritance_paths
        )

        return {
            **os.environ,
            **custom_vars,
        }

    def get_python_cmd(
        self,
        executable: Path,
        connection: ConnectionType,
    ) -> List[Union[str, Path]]:
        """Return the command to run the agent process with."""
        raise NotImplementedError

    def handle_agent_log(self, line: str, level: LogLevel) -> None:
        """Handle a log line emitted by the agent process. The level will be either
        STDOUT or STDERR."""
        raise NotImplementedError



from __future__ import annotations

from contextlib import contextmanager
from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Generic,
    Iterator,
    Optional,
    TypeVar,
)

from isolate.backends.settings import DEFAULT_SETTINGS, IsolateSettings
from isolate.logs import Log, LogLevel, LogSource

__all__ = [
    "BasicCallable",
    "CallResultType",
    "EnvironmentConnection",
    "BaseEnvironment",
    "EnvironmentCreationError",
]

ConnectionKeyType = TypeVar("ConnectionKeyType")
CallResultType = TypeVar("CallResultType")
BasicCallable = Callable[[], CallResultType]


class EnvironmentCreationError(Exception):
    """Raised when the environment cannot be created."""


class BaseEnvironment(Generic[ConnectionKeyType]):
    """Represents a managed environment definition for an isolatation backend
    that can be used to run Python code with different set of dependencies."""

    BACKEND_NAME: ClassVar[Optional[str]] = None

    settings: IsolateSettings = DEFAULT_SETTINGS

    @classmethod
    def from_config(
        cls,
        config: Dict[str, Any],
        settings: IsolateSettings = DEFAULT_SETTINGS,
    ) -> BaseEnvironment:
        """Create a new environment from the given configuration."""
        raise NotImplementedError

    @property
    def key(self) -> str:
        """A unique identifier for this environment (combination of requirements,
        python version and other relevant information) that can be used for caching
        and identification purposes."""
        raise NotImplementedError

    def create(self) -> ConnectionKeyType:
        """Setup the given environment, and return all the information needed
        for establishing a connection to it."""
        raise NotImplementedError

    def destroy(self, connection_key: ConnectionKeyType) -> None:
        """Dismantle this environment. Might raise an exception if the environment
        does not exist."""
        raise NotImplementedError

    def exists(self) -> bool:
        """Return True if the environment already exists."""
        raise NotImplementedError

    def open_connection(
        self, connection_key: ConnectionKeyType
    ) -> EnvironmentConnection:
        """Return a new connection to the environment with using the `connection_key`."""
        raise NotImplementedError

    @contextmanager
    def connect(self) -> Iterator[EnvironmentConnection]:
        """Create the given environment (if it already doesn't exist) and establish a connection
        to it."""
        connection_key = self.create()
        with self.open_connection(connection_key) as connection:
            yield connection

    def apply_settings(self, settings: IsolateSettings) -> None:
        """Apply the new settings to this environment."""
        self.settings = settings

    def log(
        self,
        message: str,
        *,
        level: LogLevel = LogLevel.DEBUG,
        source: LogSource = LogSource.BUILDER,
    ) -> None:
        """Log a message."""
        log_msg = Log(message, level=level, source=source, bound_env=self)
        self.settings.log(log_msg)


@dataclass
class EnvironmentConnection:
    environment: BaseEnvironment

    def __enter__(self) -> EnvironmentConnection:
        return self

    def __exit__(self, *exc_info):
        return None

    def run(
        self,
        executable: BasicCallable,
        *args: Any,
        **kwargs: Any,
    ) -> CallResultType:
        """Run the given executable inside the environment, and return the result.
        If the executable raises an exception, then it will be raised directly."""
        raise NotImplementedError

    def log(
        self,
        message: str,
        *,
        level: LogLevel = LogLevel.TRACE,
        source: LogSource = LogSource.BRIDGE,
    ) -> None:
        """Log a message through the bound environment."""
        self.environment.log(message, level=level, source=source)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['isolate',
 'isolate.backends',
 'isolate.connections',
 'isolate.connections._local',
 'isolate.connections.grpc',
 'isolate.connections.grpc.definitions',
 'isolate.connections.ipc',
 'isolate.server',
 'isolate.server.definitions']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'grpcio>=1.49',
 'importlib-metadata>=4.4',
 'protobuf',
 'rich>=12.0',
 'tblib>=1.7.0,<2.0.0',
 'virtualenv>=20.4']

entry_points = \
{'isolate.backends': ['conda = isolate.backends.conda:CondaEnvironment',
                      'isolate-server = isolate.backends.remote:IsolateServer',
                      'local = isolate.backends.local:LocalPythonEnvironment',
                      'pyenv = isolate.backends.pyenv:PyenvEnvironment',
                      'virtualenv = '
                      'isolate.backends.virtualenv:VirtualPythonEnvironment']}

setup_kwargs = {
    'name': 'isolate',
    'version': '0.7.2',
    'description': 'Managed isolated environments for Python',
    'long_description': '# Isolate\n\n> :warning: **Isolate** is still very young, and none of the APIs should be considered stable.\n\nRun any Python function, with any dependencies, in any machine you want. Isolate offers a\npluggable end-to-end solution for building, managing, and using isolated environments (virtualenv,\nconda, remote, and more).\n\n## Try it!\n\n```py\nfrom isolate import Template, LocalBox\n\n# Build you first environment by specifying its kind (like virtualenv or conda)\ntemplate = Template("virtualenv")\n\n# Add some packages to it.\ntemplate << "pyjokes==0.5.0"\n\n# Forward it to a box (your local machine, or a remote machine)\nenvironment = template >> LocalBox()\n\n# And then, finally try executing some code\n\ndef get_pyjokes_version():\n    import pyjokes\n\n    return pyjokes.__version__\n\n# If pyjokes==0.6.0 is installed in your local environment, it is going to print\n# 0.6.0 here.\nprint("Installed pyjokes version: ", get_pyjokes_version())\n\n# But if you run the same function in an isolated environment, you\'ll get\n# 0.5.0.\nprint("Isolated pyjokes version: ", environment.run(get_pyjokes_version))\n```\n\n## Motivation\n\n![XKCD 1987](https://imgs.xkcd.com/comics/python_environment.png)\n\nThe fact that nearly every piece of software uses some other libraries or some\nother programs is undeniable. Each of these come with their set of dependencies,\nand this chain moves forward. Once there are enough \'nodes\' in the chain, then\nthe ["dependency mess"](https://en.wikipedia.org/wiki/Dependency_hell) starts\nto surface and our lives become much harder.\n\nPython tried to solve it by recommending the "virtual environment" concept. In\ntheory it was designed to isolate environments of different projects, so my project\nA can depend on `pandas==1.0.0` while B depends on `pandas==2.0.0` and whichever\nproject I choose to work with, I just activate its own environment.\n\nOverall this was a very nice solution that did work, and still continues to work\nfor this use case. But as with every other scoped fix, in time other problems started\nto appear that demand a much narrower scope (like defining module-level dependencies,\nor even function-level ones for cloud runtimes that allow seamless integration with the\nrest of your code running in a different machine).\n\nHowever, unlike "virtual environment" concept, each of the projects that tried to tackle\nthis problem lacked a universal interface which one can simply define a set of requirements\n(this might be dependencies, size of the machine that is needed to run it, or something completely\ndifferent) and can change it without any loss. Isolate is working towards a future where this\ntransititon is as seamless as the transition from your local environment to the remote\nenvironment.\n',
    'author': 'Features & Labels',
    'author_email': 'hello@fal.ai',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Any, ClassVar, Dict

from isolate.backends import BaseEnvironment
from isolate.backends.common import sha256_digest_of
from isolate.backends.settings import DEFAULT_SETTINGS, IsolateSettings
from isolate.connections import PythonIPC


@dataclass
class LocalPythonEnvironment(BaseEnvironment[Path]):
    BACKEND_NAME: ClassVar[str] = "local"

    @classmethod
    def from_config(
        cls,
        config: Dict[str, Any],
        settings: IsolateSettings = DEFAULT_SETTINGS,
    ) -> BaseEnvironment:
        environment = cls(**config)
        environment.apply_settings(settings)
        return environment

    @property
    def key(self) -> str:
        return sha256_digest_of(sys.exec_prefix)

    def create(self) -> Path:
        return Path(sys.exec_prefix)

    def destroy(self, connection_key: Path) -> None:
        raise NotImplementedError("LocalPythonEnvironment cannot be destroyed")

    def exists(self) -> bool:
        return True

    def open_connection(self, connection_key: Path) -> PythonIPC:
        return PythonIPC(self, connection_key)



# agent-requires: isolate[server]

from __future__ import annotations

import traceback
from argparse import ArgumentParser
from concurrent import futures
from dataclasses import dataclass
from typing import Iterator, cast

import grpc
from grpc import ServicerContext, StatusCode

from isolate.connections.common import SerializationError, serialize_object
from isolate.connections.grpc import definitions
from isolate.connections.grpc.interface import from_grpc, to_grpc
from isolate.logs import Log, LogLevel, LogSource


@dataclass
class AgentServicer(definitions.AgentServicer):
    def Run(
        self,
        request: definitions.SerializedObject,
        context: ServicerContext,
    ) -> Iterator[definitions.PartialRunResult]:
        yield from self.log(f"A connection has been established: {context.peer()}!")

        if request.was_it_raised:
            return self.abort_with_msg(
                "The input function must be callable, not a raised exception.",
                context,
            )

        try:
            # TODO: technically any sort of exception could be raised here, since
            # depickling is basically involves code execution from the *user*.
            function = from_grpc(request)
        except SerializationError:
            yield from self.log(traceback.format_exc())
            return self.abort_with_msg(
                "The input function could not be deserialized.",
                context,
            )

        if not callable(function):
            return self.abort_with_msg(
                f"The input function must be callable, not {type(function).__name__}.",
                context,
            )

        yield from self.log("Starting the execution of the input function.")

        was_it_raised = False
        stringized_tb = None
        try:
            result = function()
        except BaseException as exc:
            result = exc
            was_it_raised = True
            num_frames = len(traceback.extract_stack()[:-5])
            stringized_tb = "".join(traceback.format_exc(limit=-num_frames))

        yield from self.log("Completed the execution of the input function.")

        try:
            definition = serialize_object(request.method, result)
        except SerializationError:
            yield from self.log(traceback.format_exc(), level=LogLevel.ERROR)
            return self.abort_with_msg(
                "The result of the input function could not be serialized.",
                context,
            )
        except BaseException:
            yield from self.log(traceback.format_exc(), level=LogLevel.ERROR)
            return self.abort_with_msg(
                "An unexpected error occurred while serializing the result.", context
            )

        yield from self.log("Sending the result.")
        serialized_obj = definitions.SerializedObject(
            method=request.method,
            definition=definition,
            was_it_raised=was_it_raised,
            stringized_traceback=stringized_tb,
        )
        yield definitions.PartialRunResult(
            result=serialized_obj,
            is_complete=True,
            logs=[],
        )

    def log(
        self,
        message: str,
        level: LogLevel = LogLevel.TRACE,
        source: LogSource = LogSource.BRIDGE,
    ) -> Iterator[definitions.PartialRunResult]:
        log = to_grpc(Log(message, level=level, source=source))
        log = cast(definitions.Log, log)
        yield definitions.PartialRunResult(result=None, is_complete=False, logs=[log])

    def abort_with_msg(
        self,
        message: str,
        context: ServicerContext,
        *,
        code: StatusCode = StatusCode.INVALID_ARGUMENT,
    ) -> None:
        context.set_code(code)
        context.set_details(message)
        return None


def create_server(address: str) -> grpc.Server:
    """Create a new (temporary) gRPC server listening on the given
    address."""
    server = grpc.server(
        futures.ThreadPoolExecutor(max_workers=1),
        maximum_concurrent_rpcs=1,
    )

    # Local server credentials allow us to ensure that the
    # connection is established by a local process.
    server_credentials = grpc.local_server_credentials()
    server.add_secure_port(address, server_credentials)
    return server


def run_agent(address: str) -> int:
    """Run the agent servicer on the given address."""
    server = create_server(address)
    servicer = AgentServicer()

    # This function just calls some methods on the server
    # and register a generic handler for the bridge. It does
    # not have any global side effects.
    definitions.register_agent(servicer, server)

    server.start()
    server.wait_for_termination()
    return 0


def main() -> int:
    parser = ArgumentParser()
    parser.add_argument("address", type=str)

    options = parser.parse_args()
    return run_agent(options.address)


if __name__ == "__main__":
    main()



import socket
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import Any, ContextManager, Iterator, List, Tuple, Union, cast

import grpc

from isolate.backends import (
    BasicCallable,
    CallResultType,
    EnvironmentConnection,
)
from isolate.connections._local import PythonExecutionBase, agent_startup
from isolate.connections.common import serialize_object
from isolate.connections.grpc import agent, definitions
from isolate.connections.grpc.interface import from_grpc, to_grpc
from isolate.logs import Log, LogLevel, LogSource


class AgentError(Exception):
    """An internal problem caused by (most probably) the agent."""


@dataclass
class GRPCExecutionBase(EnvironmentConnection):
    """A customizable gRPC-based execution backend."""

    def start_agent(self) -> ContextManager[Tuple[str, grpc.ChannelCredentials]]:
        """Starts the gRPC agent and returns the address it is listening on and
        the required credentials to connect to it."""
        raise NotImplementedError

    def _run_through_grpc(
        self,
        function: definitions.SerializedObject,
        *,
        max_wait_timeout: float = 10,
    ) -> Iterator[definitions.PartialRunResult]:
        """Starts a new agent process, sends the given 'function' to the agent and
        awaits until the result is received. This function takes and returns raw gRPC
        messages (use `run()` to actually execute Python functions and receive Python
        objects).

        If the gRPC server is not accessible at the passed maximum timeout, an
        AgentError will be raised."""

        # Implementation details
        # ======================
        #
        #  RPC Flow:
        #  ---------
        #  1. [controller]: Spawn the agent.
        #  2.      [agent]: Start listening at the given address.
        #  3. [controller]: Await *at most* max_wait_timeout seconds for the agent to be available
        #                   if it doesn't do it until then, raise an AgentError.
        #  4. [controller]: If the server is available, then establish the bridge and pass the
        #                   'function' as the input.
        #  5.      [agent]: Receive the function, deserialize it, start the execution.
        #  6. [controller]: Watch agent for logs (stdout/stderr), and as soon as they appear
        #                   call the log handler.
        #  7.      [agent]: Once the execution of the function is finished, send the result
        #                   using the same serialization method.
        #  8. [controller]: Receive the result back and return it.

        with self.start_agent() as (address, credentials):
            with grpc.secure_channel(address, credentials) as channel:
                channel_status = grpc.channel_ready_future(channel)
                try:
                    channel_status.result(timeout=max_wait_timeout)
                except grpc.FutureTimeoutError:
                    raise AgentError(
                        f"Couldn't connect to the gRPC server in the agent (listening at {address}) "
                        "in time."
                    )

                agent = definitions.AgentStub(channel)
                yield from agent.Run(function)

    def run(
        self,
        executable: BasicCallable,
        *args: Any,
        **kwargs: Any,
    ) -> CallResultType:
        method = self.environment.settings.serialization_method
        function = definitions.SerializedObject(
            method=method,
            definition=serialize_object(method, executable),
            was_it_raised=False,
            stringized_traceback=None,
        )

        for partial_result in self._run_through_grpc(function):
            for raw_log in partial_result.logs:
                log = from_grpc(raw_log)
                self.log(log.message, level=log.level, source=log.source)

            if partial_result.is_complete:
                if not partial_result.result:
                    raise AgentError(
                        "The agent didn't return a result, but it should have."
                    )

                return cast(CallResultType, from_grpc(partial_result.result))

        raise AgentError(
            "No result object was received from the agent (it never set is_complete to True)."
        )


class LocalPythonGRPC(PythonExecutionBase[str], GRPCExecutionBase):
    @contextmanager
    def start_agent(self) -> Iterator[Tuple[str, grpc.ChannelCredentials]]:
        def find_free_port() -> Tuple[str, int]:
            """Find a free port in the system."""
            with socket.socket() as _temp_socket:
                _temp_socket.bind(("", 0))
                return _temp_socket.getsockname()

        host, port = find_free_port()
        address = f"{host}:{port}"
        process = None
        try:
            with self.start_process(address) as process:
                yield address, grpc.local_channel_credentials()
        finally:
            if process is not None:
                # TODO: should we check the status code here?
                process.terminate()

    def get_python_cmd(
        self,
        executable: Path,
        connection: str,
    ) -> List[Union[str, Path]]:
        return [
            executable,
            agent_startup.__file__,
            agent.__file__,
            connection,
        ]

    def handle_agent_log(self, line: str, level: LogLevel) -> None:
        self.log(line, level=level, source=LogSource.USER)



from __future__ import annotations

import shutil
import tempfile
from contextlib import contextmanager
from dataclasses import dataclass, field, replace
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, Callable, Dict, Iterator, NewType, Optional

from platformdirs import user_cache_dir

if TYPE_CHECKING:
    from isolate.backends import BaseEnvironment

_SYSTEM_TEMP_DIR = Path(tempfile.gettempdir())


class LogSource(str, Enum):
    """Represents where the log orinates from."""

    # During the environment creation process (e.g. if the environment
    # is already created/cached, then no logs from this source will be
    # emitted).
    BUILDER = "builder"

    # During the environment execution process (from the server<->agent
    # communication, mostly for debugging purposes).
    BRIDGE = "bridge"

    # From the user script itself (e.g. a print() call in the given
    # function). The stream will be attached as level (stdout or stderr)
    USER = "user"


class LogLevel(str, Enum):
    """Represents the log level."""

    TRACE = "trace"
    DEBUG = "debug"
    INFO = "info"
    WARNING = "warning"
    ERROR = "error"

    # For user scripts
    STDOUT = "stdout"
    STDERR = "stderr"


@dataclass
class Log:
    """A structured log message with an option source and level."""

    message: str
    source: LogSource
    level: LogLevel = LogLevel.INFO
    bound_env: Optional[BaseEnvironment] = field(default=None, repr=False)

    def __str__(self) -> str:
        parts = []
        if self.bound_env:
            parts.append(f"[{self.bound_env.key[:6]}]")
        else:
            parts.append("[global]")

        parts.append(f"[{self.source}]".ljust(10))
        parts.append(f"[{self.level}]".ljust(10))
        return " ".join(parts) + self.message
